// Filename: BookCatalog.cpp
// S.Chan, Jul 17, 2005

#include "stdafx.h"
#include "BookCatalog.h"

#define MAX_BLANKS    1000

using namespace std;

//////////////////////////////////////////////////////////////////////////////

// String constants used for parsing comparisons to expected element names, etc.

// Element names.
static const string ELEMENT_CATALOG     = "catalog";
static const string ELEMENT_BOOK        = "book";
static const string ELEMENT_AUTHOR      = "author";
static const string ELEMENT_TITLE       = "title";
static const string ELEMENT_GENRE       = "genre";
static const string ELEMENT_PRICE       = "price";
static const string ELEMENT_PUBLISHDATE = "publish_date";
static const string ELEMENT_DESCRIPTION = "description";

// Attribute names.
static const string ATTRIBUTE_ID = "id";

// CBook /////////////////////////////////////////////////////////////////////

CBook::CBook()
{
    m_bookId = "";
    m_author = "";
    m_title = "";
    m_genre = "";
    m_price = "";
    m_publishDate = "";
    m_description = "";
}

CBook::~CBook()
{
}

std::string CBook::ToString() const
{
    string result = "";

    char buf[1000];
    sprintf(buf, "BookID:\t%s\n", m_bookId.c_str());
    result += string(buf);
    sprintf(buf, "Author:\t%s\n", m_author.c_str());
    result += string(buf);
    sprintf(buf, "Title: \t%s\n", m_title.c_str());
    result += string(buf);
    sprintf(buf, "Genre: \t%s\n", m_genre.c_str());
    result += string(buf);
    sprintf(buf, "Price: \t%s\n", m_price.c_str());
    result += string(buf);
    sprintf(buf, "Published: %s\n", m_publishDate.c_str());
    result += string(buf);
    sprintf(buf, "Description: %s\n", m_description.c_str());
    result += string(buf);

    return result;
}

// CBookCatalog //////////////////////////////////////////////////////////////

CBookCatalog::CBookCatalog()
{
    m_foundCatalog = false;
    m_currBook = NULL;
    m_currBookProperty = CBook::PROPERTY_NONE;

    // Create XML parser and attach ourself as
    // the XML events/element handler.
    m_xmlParser = new CXmlParser;
    m_xmlParser->AttachElementHandler(this);
}

CBookCatalog::~CBookCatalog()
{
    DeleteBooks();

    delete m_xmlParser;
}

void CBookCatalog::OnXmlStartElement(const CXmlElement& xmlElement)
{
    if ( m_foundCatalog )
    {
        const string& elemName = xmlElement.GetElementName();

        if ( m_currBook == NULL )
        {
            // We expect a "book" element.
            if ( elemName == ELEMENT_BOOK )
            {
                // Create a new CBook instance.
                string bookId = xmlElement.GetAttributeValue(ATTRIBUTE_ID);
                if ( !bookId.empty() )
                {
                    m_currBook = new CBook;
                    m_currBook->SetBookId(bookId);
                }
            }
        }
        else
        {
            // Set the current book property being parsed.
            if ( elemName == ELEMENT_AUTHOR )
                m_currBookProperty = CBook::PROPERTY_AUTHOR;
            else if ( elemName == ELEMENT_TITLE )
                m_currBookProperty = CBook::PROPERTY_TITLE;
            else if ( elemName == ELEMENT_GENRE )
                m_currBookProperty = CBook::PROPERTY_GENRE;
            else if ( elemName == ELEMENT_PRICE )
                m_currBookProperty = CBook::PROPERTY_PRICE;
            else if ( elemName == ELEMENT_PUBLISHDATE )
                m_currBookProperty = CBook::PROPERTY_PUBLISHDATE;
            else if ( elemName == ELEMENT_DESCRIPTION )
                m_currBookProperty = CBook::PROPERTY_DESCRIPTION;
            else
                m_currBookProperty = CBook::PROPERTY_NONE;
        }
    }
    else
    {
        // We expect a "catalog" element.
        if ( xmlElement.GetElementName() == ELEMENT_CATALOG )
        {
            m_foundCatalog = true;
        }
    }
}

void CBookCatalog::OnXmlElementData(const std::string& elementData, int depth)
{
    if ( !m_foundCatalog )
        return;
    if ( m_currBook == NULL )
        return;

    switch( m_currBookProperty )
    {
    case CBook::PROPERTY_AUTHOR:
        m_currBook->SetAuthor(m_currBook->GetAuthor() + elementData);
        break;
    case CBook::PROPERTY_TITLE:
        m_currBook->SetTitle(m_currBook->GetTitle() + elementData);
        break;
    case CBook::PROPERTY_GENRE:
        m_currBook->SetGenre(m_currBook->GetGenre() + elementData);
        break;
    case CBook::PROPERTY_PRICE:
        m_currBook->SetPrice(m_currBook->GetPrice() + elementData);
        break;
    case CBook::PROPERTY_PUBLISHDATE:
        m_currBook->SetPublishDate(m_currBook->GetPublishDate() + elementData);
        break;
    case CBook::PROPERTY_DESCRIPTION:
        if ( !m_currBook->GetDescription().empty() )
            m_currBook->SetDescription(m_currBook->GetDescription() + "\n" + elementData);
        else
            m_currBook->SetDescription(elementData);
        break;
    default:
        break;
    }
}

void CBookCatalog::OnXmlEndElement(const CXmlElement& xmlElement)
{
    if ( !m_foundCatalog )
        return;

    if ( m_currBook != NULL )
    {
        const string& elemName = xmlElement.GetElementName();

        // Check if we are at the end of a book.
        if ( elemName == ELEMENT_BOOK )
        {
            // Add the completed book to the Catalog but check for duplicates.
            map<string,CBook*>::iterator result = m_books.find(m_currBook->GetBookId());
            if ( result == m_books.end() )
            {
                // Insert new book.
                m_books.insert(make_pair(m_currBook->GetBookId(), m_currBook));
            }
            else
            {
                // Update existing book.
                CBook* pExistingBook = result->second;
                *pExistingBook = *m_currBook;
            }

            // Reset current book.
            m_currBook = NULL;
            return;
        }
    }

    m_currBookProperty = CBook::PROPERTY_NONE;
}

void CBookCatalog::OnXmlError(int line, int column, const std::string& errorText, unsigned long errorCode)
{
    // Ignore errors. We just find whatever books there are in the XML.
}

bool CBookCatalog::OnXmlAbortParse(const CXmlElement& xmlElement)
{
    return false;
}

bool CBookCatalog::BuildFromXml(const std::string& xmlPath)
{
    if ( !m_xmlParser->IsReady() )
        return false;

    // Do the parsing.
    m_foundCatalog = false;
    m_currBook = NULL;
    m_currBookProperty = CBook::PROPERTY_NONE;
    bool result = m_xmlParser->Parse(xmlPath);

    return result;
}

const CBook* CBookCatalog::FindBook(const std::string& bookId) const
{
    if ( bookId.empty() )
        return NULL;

    map<string,CBook*>::const_iterator result = m_books.find(bookId);
    if ( result == m_books.end() )
        return NULL;

    return result->second;
}

int CBookCatalog::GetBookCount() const
{
    return m_books.size();
}

void CBookCatalog::DeleteBooks()
{
    m_foundCatalog = false;
    m_currBook = NULL;
    m_currBookProperty = CBook::PROPERTY_NONE;

    // Delete book objects.
    map<string,CBook*>::iterator i = m_books.begin();
    for( ; i != m_books.end(); ++i)
    {
        CBook* pBook = i->second;
        delete pBook;
    }
    m_books.clear();
}

// END
